#include "Matrix.h"

/* --- ReadEquations inputs and builds the augmented matrix.

   Receive: AugMat, an (empty) matrix.
   Input: The linear system coefficients, and constant values.
   Return: AugMat, filled with the input values.
----------------------------------------------------------------*/

void ReadEquations(Matrix& AugMat)
{
  int
    NumEquations;

  for (;;)                           // get matrix dimensions
  {
      cout << "\nPlease enter the number of equations in the system: ";
      cin >> NumEquations;
      if (NumEquations > 1)
         break;
      cerr << "\n*** At least two equations are needed ...\n";
  }
                                      // build a matrix of
  AugMat = Matrix(NumEquations,       // NumEquations rows &
                  NumEquations+1);    // NumEquations+1 columns

                                      // get the coefficients
  cout << "\nPlease enter the coefficient matrix row-by-row...\n";
  for (int r = 0; r < NumEquations; r++)
    for (int c = 0; c < NumEquations; c++)
      cin >> AugMat[r][c];
                                      // get the constants
  cout << "\nPlease enter the constant vector...\n";
  for (r = 0; r < NumEquations; r++)
    cin >> AugMat[r][NumEquations];
}

/* --- Abs performs the absolute value function on a double value --- */

double Abs(double Val)
{
  return (Val < 0) ? -(Val) : Val;
}
 
/* --- Swap exchanges the values of two double variables ------------- */

void Swap(double& a, double& b)
{
  double
    t = a;

  a = b;
  b = t;
}

/* --- Reduce simplifies an augmented matrix into triangular form ------

   Receive: AugMat, and augmented matrix.
   Return:  if the linear system is non-singular:
               AugMat, reduced to a triangular matrix; and
               True.
            else
               False.
------------------------------------------------------------------------*/

int Reduce(Matrix& AugMat)
{
  const double
    Epsilon = 1.0E-6;
  int
    Singular = 0,
    i = 0,
    j,
    k,
    n = AugMat.NumberOfRows(),
    PivotRow;
  double
    Mult,
    AbsPivot;

  while ((!Singular) && (i < n))
  {
     AbsPivot = Abs(AugMat[i][i]);
     PivotRow = i;

     for (k = i+1; k < n; k++)
        if (Abs(AugMat[k][i]) > AbsPivot)
        {
            AbsPivot = Abs(AugMat[k][i]);
            PivotRow = k;
        }

     Singular = AbsPivot < Epsilon;

     if (!Singular)
     {
         if (i != PivotRow)
            for (j = 0; j <= n; j++)
              Swap(AugMat[i][j], AugMat[PivotRow][j]);
 
         for (j = i+1; j < n; j++)
         {
              Mult = -AugMat[j][i] / AugMat[i][i];
              for (k = i; k <= n; k++)
                AugMat[j][k] = AugMat[j][k] + Mult * AugMat[i][k];
         }
     }
     i++;
  }
  return Singular;
}

/* --- Solve determines the solution of the linear system -------------

   Receive: AugMat, an augmented matrix (reduced to triangular form).
   Return: The solution vector to the linear system.
----------------------------------------------------------------------*/

Matrix Solve(Matrix& AugMat)
{
  Matrix
    SolVect(1, AugMat.NumberOfRows());
  int
    n = AugMat.NumberOfRows()-1;

  SolVect[0][n] = AugMat[n][n+1] / AugMat[n][n];

  for (int i = n-1; i >= 0; i--)
    {
      SolVect[0][i] = AugMat[i][n+1];

      for (int j = i+1; j <= n; j++)
        SolVect[0][i] -= AugMat[i][j] * SolVect[0][j];

      SolVect[0][i] /= AugMat[i][i];
    }

  return SolVect;
}
