/* This program reads employee numbers and hourly rates from the file
   "emp.dat", hours worked from the keyboard, and calculates employee
   wages, writing them and the total of all wages to the file
   "payroll.dat".

   Input (emp.dat):      A sequence of lines of the form:

                              employee-id-number   hourly-rate

   Input (keyboard):     For each employee, the number of hours worked

   Output (payroll.dat): Column headings, and a series of lines with:

                              employee-id-number   wages

                         in an easy-to-read format; also, the total
                         wages for all employees
---------------------------------------------------------------------*/

#include <iostream.h>                         // for terminal I/O
#include <fstream.h>                          // for file I/O
#include <iomanip.h>                          // for formatted I/O
#include <stdlib.h>                           // for exit()

void GetEmployeeInfo(fstream&, long&, double&);

int main(void)
{
   const char
      InFile[] = "emp.dat",                  // name of input file
      OutFile[] = "payroll.dat",             // name of output file
                                             // error messages
      OpenFailedMsg[] = "\n*** main: unable to open file: ";

   cout << "\nThis program gets data from the user and from a file"
        << "\n\t\'" << InFile
        << "', computes each employee's pay, and displays"
        << "\n\tthis information and the total of all wages in a "
        << "\n\treadable format in the file '"
        << OutFile << "'.\n\n";

   fstream
      InStream(InFile, ios::in),             // stream to input file
      OutStream(OutFile, ios::out);          // stream to output file

   if (InStream.fail())                      // check for
   {                                         //   successful opens
      cerr << OpenFailedMsg << InFile << endl;
      exit (-1);
   }
   else if (OutStream.fail())
   {
      cerr << OpenFailedMsg << OutFile << endl;
      exit (-1);
   }

   long
      EmpIdNumber;                           // employee's ID number
   double
      Pay,                                   // employee's pay
      TotalPay = 0.0;                        // running pay-total

                                             // write column headings
   OutStream << "\n\t" << "  Employee  " << '\t' << "     Pay     "
             << "\n\t" << "============" << '\t' << "============="
             << endl;

   OutStream << setprecision(2)              // 2 dec. places
             << setiosflags(ios::showpoint | ios::fixed);
                                             // fixed, not scientific

   GetEmployeeInfo(InStream, EmpIdNumber, Pay);
                                             // get 1st set of info
   while (!InStream.eof())                   // while more values left,
   {
      TotalPay += Pay;                       //  keep running total
                                             //  format the lines
      OutStream << "\t "
                << EmpIdNumber / 1000000 << '-'
                << EmpIdNumber % 1000000 / 10000 << '-'
                << EmpIdNumber % 10000
                << '\t'
                << setw(9) << Pay
                << endl;
                                             //  get next set of info
      GetEmployeeInfo(InStream, EmpIdNumber, Pay);
   }                                         // end while

   InStream.close();                         // close the input file

                                             // output summary info
   OutStream << "\n\nThe total payroll comes to $"
             << TotalPay << "\n\n";

   OutStream.close();                        // close the output file

   cout
     << "\nDone. Results are in the file '"  // and tell user where
     << OutFile << "'.\n\n";                 // to find the results

   return 0;
}


/*-----------------------------------------------------------------
ComputeWages determines the pay for a rate and time,
   including overtime.

   Receive:  HourlyRate, the hourly rate for an employee;
             Hours, the number of hours that employee worked
   Return:   The pay earned by that employee
------------------------------------------------------------------*/

double ComputeWages(double HourlyRate, double Hours)
{
   const double
      OverTimeFactor = 1.5,                 // bonus for OT hours
      RegularLimit = 40.0;                  // cutoff for OT
   double
      RegularWages,                         // non-OT earnings
      OverTimeWages = 0.0;                  // OT earnings

   if (Hours <= RegularLimit)               // if no OT,
      RegularWages = HourlyRate * Hours;    //   just compute non-OT
   else                                     // otherwise
   {                                        //   compute OT & non-OT
      RegularWages = HourlyRate * RegularLimit;
      OverTimeWages = HourlyRate * OverTimeFactor
                    * (Hours - RegularLimit);
   }
   return RegularWages + OverTimeWages;     // return non-OT + OT
}


/*------------------------------------------------------------------
GetEmployeeInfo inputs all necessary information for an employee, 
   and uses ComputeWages to compute the amount that employee is 
   to be paid.

   Receive:          InS, an fstream to a file containing 
                       employee data
   Input(InS):       An employee Id number (EmpId) and HourlyRate.
   Output(Screen):   A prompt for keyboard input
   Input(keyboard):  HoursWorked by the employee
   Return:           EmpId and the Pay earned by the employee
------------------------------------------------------------------*/

void GetEmployeeInfo(fstream& InS, long& EmpId, double& Pay)
{
   double
      HoursWorked,
      HourlyRate;

   InS >> EmpId >> HourlyRate;              // get emp# and pay rate

   if (!InS.eof())                          // if there was data left
   {                                        //   get hours worked
      cout << "\nHow many hours did employee "
           << EmpId << " work? ";
      cin >> HoursWorked;
                                            //   calculate wages
      Pay = ComputeWages(HourlyRate, HoursWorked);
   }
   else
      Pay = 0.0;                            // use when end of file
}
